# ----------------------------------------------------------------------
#  APP: Octa-View - GUI for simple Octave equations
#
#  This program lets you type in some Octave code and evaluate it.
#  Errors in the program are highlights, and results are plotted.
# ======================================================================
#  AUTHOR:  Michael McLennan, Purdue University
#  Copyright (c) 2011  Purdue University
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
package require Itcl
package require BLT
package require RapptureGUI
package require Img

wm title . "OCTAViEw"
wm geometry . 700x600
wm protocol . WM_DELETE_WINDOW exit
wm withdraw .

set InstallDir [file dirname [file normalize [info script]]]
lappend auto_path $InstallDir

Rappture::icon::searchpath [file join $InstallDir images]

set CurrentDir $env(HOME)

option add *Text.background white startupFile
option add *Tabset.borderWidth 0 startupFile
option add *XyResult.width 1i startupFile
option add *XyResult.height 1i startupFile
option add *in.ctrl*background gray startupFile
option add *in.ctrl*highlightBackground gray startupFile
option add *in.ctrl*Button.relief flat startupFile
option add *in.ctrl*Button.overRelief raised startupFile
option add *in.ctrl*Button.padX 4 startupFile
option add *run.cntl.go.font {Helvetica -18 bold italic} startupFile
option add *run.cntl.go.padX 4 startupFile

# tweak the style sheet for the HTML viewer
itcl::body Rappture::HTMLviewer::_config {args} {
    component html style -id "author" "
      body {
        background: $itk_option(-background);
        color: $itk_option(-foreground);
        font: 12px helvetica,arial;
      }
      ul {
        list-style-image: url(bullet.png);
      }
      li {
        line-height: 24px;
      }
      a {
        color: $itk_option(-linknormalcolor);
        text-decoration: underline;
      }
      a:hover {
        color: $itk_option(-linkactivecolor);
        background: $itk_option(-linkactivebackground);
      }
    "
}

# plug in Rappture's error handler
Rappture::bugreport::install
toplevel .main  ;# bug report stuff looks for .main
wm withdraw .main

# use Rappture's grab stack
Rappture::grab::init

# ----------------------------------------------------------------------
# USAGE: get_octave_file 
#
#	Read the parameters.hz file and get the name of the file
#	listed.  We don't care if the file parameter has a name 
#	(e.g. file(name):path) or if there are other parameters 
#       listed.  The first file parameter wins.
#
# ----------------------------------------------------------------------
proc get_octave_file { fname } {
    set fid [open $fname r]
    set contents [read $fid]
    close $fid

    # Pick only the first file entry.  Maybe someday we'll handle more than
    # one source file (in separate tabs maybe).
    
    set lines [split $contents "\n"]
    foreach line $lines {
	if { [string compare -length 4 "file" $line] != 0 } { 
	    continue
	}
	# There can't be colons in the file name, so it's OK to split
	# the line on colons.
	set parts [split $line ":"]
	set file [lindex $parts 1]

        if {[catch {
            set fid [open $file "r"]
            set contents [read $fid]
            close $fid
        } result]} {
            tk_messageBox -icon error -title "OCTAViEw: Error" -message [string totitle $result]
            return ""
        }

        return $contents
    }
    return ""
}

# ----------------------------------------------------------------------
# USAGE: main_seek_input_from_parameters
#
# Called at the start of execution to look for an input file coming
# from the tool invocation.  If the file exists, it is loaded into
# the main editor.  After too many tries, we give up and stop looking
# for the input to appear.
# ----------------------------------------------------------------------
proc main_seek_input_from_parameters {{tries 50}} {
    global env

    set fname $env(TOOL_PARAMETERS)
    if { [file readable $fname] } {
	set info [get_octave_file $fname]
	if { $info ne "" } {
	    # strip out any carriage returns that come from the Windows side
            regsub -all {\r} $info "" info

            # load the script
            .in.scrl.text delete 1.0 end
            .in.scrl.text insert end $info
            .in.scrl.text mark set insert 1.0
            .in.scrl.text see 1.0
            main_reset_results
        }
    } elseif {[incr tries -1] > 0} {
        # if we haven't tried too many times yet, keep trying
        after 200 main_seek_input_from_parameters $tries
    }
}

# ----------------------------------------------------------------------
# USAGE: main_seek_input
#
# Called at the start of execution to look for an input file coming
# from the tool invocation.  If the file exists, it is loaded into
# the main editor.  After too many tries, we give up and stop looking
# for the input to appear.
#
#
# Note: This is the deprecated method of passing parameters from the web
#	to tools.  We are supporting this only because it is still in
#	use on nanohub.org.
#
# ----------------------------------------------------------------------
proc main_seek_input {{tries 50}} {
    global env

    # look for the file in /apps/projects/workingdir/<user>/file
    set fname [file join /apps/projects/workingdir $env(USER) file]
    if {[file readable $fname]} {
        catch {
            set fid [open $fname r]
            set info [read $fid]
            close $fid

            # strip out any carriage returns that come from the Windows side
            regsub -all {\r} $info "" info

            # load the script
            .in.scrl.text delete 1.0 end
            .in.scrl.text insert end $info
            .in.scrl.text mark set insert 1.0
            .in.scrl.text see 1.0
            main_reset_results

            # done with this now -- start fresh next time
            file delete -force $fname
        }
    } elseif {[incr tries -1] > 0} {
        # if we haven't tried too many times yet, keep trying
        after 200 main_seek_input $tries
    }
}

# ----------------------------------------------------------------------
# USAGE: main_run
#
# Takes the contents from the input window and runs Octave.
# ----------------------------------------------------------------------
proc main_run {} {
    global Run

    # show the progress area, but clear it out first
    foreach win [pack slaves .right] {
        pack forget $win
    }
    pack .right.progress -expand yes -fill both
    .right.progress.scrl.text configure -state normal
    .right.progress.scrl.text delete 1.0 end
    .right.progress.scrl.text configure -state disabled

    # info area is turned off now
    set defbg [lindex [.run.cntl.info configure -background] 3]
    .run.cntl.info configure -relief flat -background $defbg

    .run.cntl.go configure -text "Abort" \
        -command "main_spinner stop; $Run abort"

    main_reset_results
    main_spinner start
    $Run clear
    $Run script [.in.scrl.text get 1.0 end]
    main_spinner stop

    # load plots into the plots tab
    set plots [$Run getplots]
    if {[llength $plots] == 0} {
        if {[catch {.right.results.tabs index -name plots} i] == 0} {
            .right.results.tabs delete $i
        }
    } else {
        if {[catch {.right.results.tabs index -name plots} i]} {
            .right.results.tabs insert 0 plots -text "Plots" \
                -window .right.results.tabs.plots -fill both
        }

        set win .right.results.tabs.plots
        $win.xy clear
        $win.which.plot choices delete 0 end

        set first ""
        foreach rec $plots {
            set pname [lindex $rec 0]
            set curvelist [lindex $rec 2]
            if {$first eq ""} { set first $pname }
            $win.which.plot choices insert end $curvelist $pname
        }
        if {$first ne ""} {
            .right.results.tabs select [.right.results.tabs index -name plots]
            $win.which.plot value $first
            main_xyplot_update
        }
    }

    # load debug info into the debug tab
    set vars [$Run getvars]
    set win .right.results.tabs.debug
    $win.which.plot value ""
    $win.which.plot choices delete 0 end
    $win.xy clear

    $win.scrl.info configure -state normal
    $win.scrl.info delete 1.0 end

    set first ""
    foreach obj $vars {
        switch -glob -- [$obj type] {
            *x1 {
                set v [$obj vector]
                $win.scrl.info insert end "[$obj name] = [$v range 0 0]\n"
            }
            *1x* {
                if {$first eq ""} { set first $obj }
                $win.which.plot choices insert end $obj [$obj name]
            }
        }
    }

    if {$first ne ""} {
        if {[llength $plots] == 0} {
            # no plots? then show debug info by default
            .right.results.tabs select [.right.results.tabs index -name debug]
        }
        $win.which.plot value [$first name]
        main_debug_update yaxis
    } elseif {[llength $vars] == 0} {
        # no variables? must have encountered an error
        .right.results.tabs select [.right.results.tabs index -name console]
        $win.which.plot value ""
    }
    $win.scrl.info configure -state disabled

    # load session text into the console area
    main_info -off  ;# puts up results area
    .right.results.tabs.console.scrl.text configure -state normal
    .right.results.tabs.console.scrl.text delete 1.0 end
    .right.results.tabs.console.scrl.text insert end [$Run getoutput]
    .right.results.tabs.console.scrl.text configure -state disabled

    # reset run button to run again
    .run.cntl.go configure -text "Run Script" -image [Rappture::icon octave-logo-sm] -command main_run
}

# ----------------------------------------------------------------------
# USAGE: main_info -showing|-on|-off|-toggle
#
# Removes the progress/results area and puts the main tutorial info
# part back up.  This is handy when you're trying to get back to the
# tutorial info.
# ----------------------------------------------------------------------
proc main_info {op} {
    global Run

    # toggle the current state for help info
    set state [expr {([.run.cntl.info cget -relief] eq "sunken")}]

    switch -exact -- $op {
        -showing { return $newstate }
        -on      { set newstate 1 }
        -off     { set newstate 0 }
        -toggle  { set newstate [expr {!$state}] }
        default  { error "bad option \"$op\": should be -on, -off, -toggle, -showing" }
    }

    # take down anything on the right and put up intro info
    foreach win [pack slaves .right] {
        pack forget $win
    }
    if {$newstate} {
        pack .right.intro -expand yes -fill both
        .run.cntl.info configure -relief sunken -background gray
    } else {
        if {[$Run isrunning]} {
            pack .right.progress -expand yes -fill both
        } else {
            pack .right.results -expand yes -fill both
        }
        set defbg [lindex [.run.cntl.info configure -background] 3]
        .run.cntl.info configure -relief flat -background $defbg
    }
}

# ----------------------------------------------------------------------
# USAGE: main_spinner start
# USAGE: main_spinner next
# USAGE: main_spinner stop
#
# Changes the icon on the "Run Scripts" button to animate a busy
# spinner while the script is running.  This lets the user know that
# the program is working -- particular if they switch to the info
# view and they're not looking at the output.
# ----------------------------------------------------------------------
proc main_spinner {op} {
    global spinner

    switch -exact -- $op {
        start {
            main_spinner stop ;# in case it is running
            set spinner(count) 0
            set spinner(callback) [after idle main_spinner next]
        }
        next {
            set img [Rappture::icon box-spinner-$spinner(count)]
            if {$img eq ""} {
                set spinner(count) 0
                set img [Rappture::icon box-spinner-$spinner(count)]
            } else {
                incr spinner(count)
            }
            .run.cntl.go configure -image $img
            set spinner(callback) [after 200 main_spinner next]
        }
        stop {
            if {[info exists spinner(callback)] && $spinner(callback) ne ""} {
                after cancel $spinner(callback)
                set spinner(callback) ""
            }
        }
        default {
            error "bad option \"$op\": should be start, next, stop"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: main_xyplot_update
#
# Invoked whenever the user changes the plot that should be displayed.
# Loads the data for the desired plot into the plotting area.  These
# are the official plots that the user requested via the plot()
# command in Octave.
# ----------------------------------------------------------------------
proc main_xyplot_update {} {
    global Run

    set win .right.results.tabs.plots

    set curvelist [$win.which.plot translate [$win.which.plot value]]
    $win.xy show $curvelist
}

# ----------------------------------------------------------------------
# USAGE: main_debug_clear
#
# Invoked whenever the user changes the plot that should be displayed
# in the "debug" tab.  The "xaxis" or "yaxis" parameter indicates
# which axis has changed.
# ----------------------------------------------------------------------
proc main_debug_clear {} {
    set win .right.results.tabs.debug
    $win.which.plot value ""
}

# ----------------------------------------------------------------------
# USAGE: main_debug_update yaxis|xaxis
#
# Invoked whenever the user changes the plot that should be displayed
# in the "debug" tab.  The "xaxis" or "yaxis" parameter indicates
# which axis has changed.
# ----------------------------------------------------------------------
proc main_debug_update {what} {
    global Run DebugCurves

    set win .right.results.tabs.debug
    switch -- $what {
        yaxis {
            set newy [$win.which.plot translate [$win.which.plot value]]
            set compatible ""

            if {$newy ne ""} {
                foreach obj [$Run getvars] {
                    if {$obj eq $newy} continue
                    if {[$obj type] eq [$newy type]} {
                        lappend compatible $obj
                    }
                }
                $win.which.versus.var choices delete 0 end
            }

            if {[llength $compatible] > 0} {
                foreach obj $compatible {
                    $win.which.versus.var choices insert end $obj [$obj name]
                }
                pack $win.which.versus -after $win.which.plot -side left -expand yes -fill x
            } else {
                pack forget $win.which.versus
            }
            $win.which.versus.var choices insert end "--" "--"
            $win.which.versus.var value "--"
            main_debug_update xaxis
        }
        xaxis {
            set newy [$win.which.plot translate [$win.which.plot value]]
            set newx [$win.which.versus.var translate [$win.which.versus.var value]]
            if {[info exists DebugCurves($newy/$newx)]} {
                set obj $DebugCurves($newy/$newx)
            } elseif {$newy ne ""} {
                if {$newx eq "--"} {
                    set obj [Curve ::#auto $newy]
                } else {
                    set obj [Curve ::#auto $newy $newx]
                }
                set DebugCurves($newy/$newx) $obj
            } else {
                set obj ""
            }
            if {$obj ne ""} {
                $win.xy show $obj
            }
        }
        default {
            error "bad option \"$what\": should be yaxis or xaxis"
        }
    }
}

# ----------------------------------------------------------------------
# USAGE: main_err <message> <line> <column>
#
# This gets called when an error is encountered while executing the
# Octave script.  Displays the error and highlights the offending
# line in the script.
# ----------------------------------------------------------------------
proc main_err {info line col} {
    .in.scrl.text tag remove err 1.0 end
    catch {.in.scrl.text delete errmarker.first errmarker.last}

    if {$line > 0} {
        .in.scrl.text tag add err $line.0 [expr {$line+1}].0
        if {$col eq "" || $col <= 0} {
            set col 0
        } else {
            # octave reports first column as column #1 but Tk uses #0
            incr col -1
        }
        set bbox [.in.scrl.text bbox $line.$col]
        if {$bbox eq ""} {
            # probably pointing to line at the very bottom
            incr line -1
            set bbox [.in.scrl.text bbox $line.$col]
            .in.scrl.text tag add err $line.0 [expr {$line+1}].0
        }

        #
        # BEWARE: Use the "errloc" mark to insert the image and tag it.
        #   Sometimes the $line.$col index overshoots the line--for
        #   example, when plotoctaview() is substituted in place of plot().
        #   Using "errloc" makes sure that the "errmarker" tag is applied
        #   correctly so the image can be deleted later.
        #
        .in.scrl.text see $line.$col
        .in.scrl.text mark set errloc $line.$col
        .in.scrl.text mark gravity errloc left
        .in.scrl.text image create errloc -image [Rappture::icon errloc]
        .in.scrl.text tag add errmarker errloc errloc+1char
        .in.scrl.text mark set insert errloc+1char

        if {$bbox ne ""} {
            foreach {x y w h} $bbox break
            set x [expr {$x+[winfo rootx .in.scrl.text]+5}]
            set y [expr {$y+[winfo rooty .in.scrl.text]+$h+5}]
            Rappture::Tooltip::cue @$x,$y $info
            .rappturetoolcue configure -icon ""
            bind .rappturetoolcue <Unmap> main_err_clear
        }
    } else {
        tk_messageBox -title "OCTAViEw: Error" -icon error -message $info
    }
}

# ----------------------------------------------------------------------
# USAGE: main_err_clear
#
# Used to clear the error indictor on the script window.  Removes
# the highlighting and the icon that shows the character nearest
# the error.
# ----------------------------------------------------------------------
proc main_err_clear {} {
    .in.scrl.text tag remove err 1.0 end
    catch {.in.scrl.text delete errmarker.first errmarker.last}
}

# ----------------------------------------------------------------------
# USAGE: main_progress <channel> <string>
#
# This gets called when output comes back from the running process.
# Shows the output in a progress window, so the user can see what's
# happening during simulation.
# ----------------------------------------------------------------------
proc main_progress {channel str} {
    .right.progress.scrl.text configure -state normal
    .right.progress.scrl.text insert end $str $channel
    .right.progress.scrl.text configure -state disabled
    .right.progress.scrl.text see end
}

# ----------------------------------------------------------------------
# USAGE: main_reset_results
#
# Used to clear the right-hand (results) side of the GUI.  Clears
# all plots and the console text.
# ----------------------------------------------------------------------
proc main_reset_results {} {
    global DebugCurves

    # clear the plots tab
    set win .right.results.tabs.plots
    $win.which.plot value ""
    $win.which.plot choices delete 0 end
    $win.xy clear

    # clear the debug tab
    set win .right.results.tabs.debug
    $win.which.plot value ""
    $win.which.plot choices delete 0 end
    $win.which.versus.var value ""
    $win.which.versus.var choices delete 0 end
    $win.xy clear

    foreach key [array names DebugCurves] {
        itcl::delete object $DebugCurves($key)
        unset DebugCurves($key)
    }

    $win.scrl.info configure -state normal
    $win.scrl.info delete 1.0 end
    $win.scrl.info configure -state disabled

    # clear the output console
    .right.results.tabs.console.scrl.text configure -state normal
    .right.results.tabs.console.scrl.text delete 1.0 end
    .right.results.tabs.console.scrl.text configure -state disabled
}

# ----------------------------------------------------------------------
# USAGE: main_file_upload
# USAGE: main_file_upload_ready path <p> data <d> error <i> ...
#
# Like main_file_open, but takes the file from the user's desktop
# via the "importfile" command.  main_file_upload kicks off the
# process.  After the file has been uploaded, main_file_upload_ready
# is called with a series of key/value pairs to complete the process
# and load the data into the app.
# ----------------------------------------------------------------------
proc main_file_upload {} {
    Rappture::filexfer::upload \
        "OCTAViEw" [list script "Octave/MATLAB Script" "Script that you want to execute in the OCTAViEw environment.  Usually a file ending with \".m\" containing Octave/MATLAB commands."] main_file_upload_ready
}

proc main_file_upload_ready {args} {
    array set info $args
    if {[info exists info(error)]} {
        if {[regsub {\n+Click anywhere.+} $info(error) "" info(error)] > 0} {
            after idle [list tk_messageBox -title "OCTAViEw: Notice" -icon info -message $info(error)]
        } else {
            after idle [list tk_messageBox -title "OCTAViEw: Error" -icon error -message $info(error)]
        }
    } else {
        # strip out any carriage returns that come from the Windows side
        regsub -all {\r} $info(data) "" info(data)

        # load the script
        .in.scrl.text delete 1.0 end
        .in.scrl.text insert end $info(data)
        .in.scrl.text mark set insert 1.0
        .in.scrl.text see 1.0
        main_reset_results
    }
}

# ----------------------------------------------------------------------
# USAGE: main_file_download
#
# Like main_file_saveas, but downloads the file to the user's desktop
# via the "exportfile" command.  This is sometimes more convenient
# saving and then downloading the file.
# ----------------------------------------------------------------------
proc main_file_download {} {
    Rappture::filexfer::download [.in.scrl.text get 1.0 end] "input.txt"
}

# ----------------------------------------------------------------------
# USAGE: main_file_open
#
# Prompts the user and loads a file from the user's hub directory.
# ----------------------------------------------------------------------
proc main_file_open {} {
    global env CurrentDir

    set fname [tk_getOpenFile -title "Open Octave Script" -parent . \
        -defaultextension .m -initialdir $CurrentDir -filetypes {
            {"Octave/MATLAB files" .m}
            {"All Files" *}
        }]

    if {$fname ne ""} {
        if {[catch {
            set fid [open $fname r]
            set info [read $fid]
            close $fid
        } result]} {
            tk_messageBox -title "OCTAViEw: Error" -icon error -message "Can't open file: $result"
            return
        }

        .in.scrl.text delete 1.0 end
        .in.scrl.text insert end $info
        .in.scrl.text mark set insert 1.0
        .in.scrl.text see 1.0

        main_reset_results

        # start in this directory next time
        set CurrentDir [file dirname $fname]
    }
}

# ----------------------------------------------------------------------
# USAGE: main_file_saveas
#
# Prompts the user to save the contents of the text area as a file
# in the home directory area.
# ----------------------------------------------------------------------
proc main_file_saveas {} {
    global env CurrentDir

    set fname [tk_getSaveFile -title "Save Octave Script" -parent . \
        -defaultextension .m -initialdir $CurrentDir -filetypes {
            {"Octave/MATLAB files" .m}
            {"All Files" *}
        }]

    if {$fname ne ""} {
        if {[catch {
            set fid [open $fname w]
            puts -nonewline $fid [.in.scrl.text get 1.0 end]
            close $fid
        } result]} {
            tk_messageBox -title "OCTAViEw: Error" -icon error -message "Can't save file: $result"
            return
        }

        # start in this directory next time
        set CurrentDir [file dirname $fname]
    }
}

# ----------------------------------------------------------------------
# OBJECT: Result
# Base class for all possible Octave results
# ----------------------------------------------------------------------
itcl::class Result {
    constructor {name type args} {
        set _name $name
        set _type $type
        eval configure $args
    }

    public method name {} { return $_name }
    public method type {} { return $_type }

    protected variable _name ""
    protected variable _type ""
}

# ----------------------------------------------------------------------
# OBJECT: VectorResult
# Represents a single vector of x values (like a BLT vector)
# ----------------------------------------------------------------------
itcl::class VectorResult {
    inherit Result

    constructor {name xvals args} {
        Result::constructor $name "double1x[llength $xvals]"
    } {
        set _vec [blt::vector create #auto]

        # is there an imaginary vector?
        if {[llength $args] >= 1} {
            set ivals [lindex $args 0]
        } else {
            set ivals ""
        }

        # any values to add?
        if {[llength $xvals] > 0} {
            if {[llength $ivals] > 0} {
                # handle extra list of complex values
                set real [blt::vector create #auto]
                eval $real append $xvals
                set imag [blt::vector create #auto]
                eval $imag append [lindex $args 0]
                $_vec expr "sqrt($real*$real + $imag*$imag)"
                blt::vector destroy $real $imag
            } else {
                # ordinary real vector
                eval $_vec append $xvals
            }
        }
    }

    destructor {
        blt::vector destroy $_vec
    }

    method vector {} { return $_vec }

    private variable _vec ""
}

# ----------------------------------------------------------------------
# OBJECT: Curve
# Like the Rappture::Curve, but interfaces with VectorResult.
# ----------------------------------------------------------------------
itcl::class Curve {
    public variable xlabel ""
    public variable ylabel ""
    public variable xscale "linear"
    public variable yscale "linear"
    public variable xlimits ""
    public variable ylimits ""

    constructor {yvec {xvec ""} args} {
        if {$xvec eq ""} {
            set _storage [VectorResult ::#auto "index" ""]
            set max [[$yvec vector] length]
            [$_storage vector] seq 1 $max $max
            set xvec $_storage
        }
        set _yvec $yvec
        set _xvec $xvec
    }
    destructor {
        if {$_storage ne ""} {
            itcl::delete object $_storage
        }
    }

    public method components {{pattern *}} { return "main" }
    public method mesh {{what -overall}} { return [$_xvec vector] }
    public method values {{what -overall}} { return [$_yvec vector] }
    public method limits {which} {
        switch -- $which {
            x - xlin { set v $_xvec; set log 0 }
            xlog { set v $_xvec; set log 1 }
            y - ylin - v - vlin { set v $_yvec; set log 0 }
            ylog - vlog { set v $_yvec; set log 1 }
            default {
                error "bad option \"$which\": should be x, xlin, xlog, y, ylin, ylog, v, vlin, vlog"
            }
        }

        blt::vector tmp zero
        [$v vector] variable vec

        if {$log} {
            # on a log scale, use abs value and ignore 0's
            [$v vector] dup tmp
            [$v vector] dup zero
            zero expr {tmp == 0}            ;# find the 0's
            tmp expr {abs(tmp)}             ;# get the abs value
            tmp expr {tmp + zero*max(tmp)}  ;# replace 0's with abs max
            set min [blt::vector expr min(tmp)]
            set max [blt::vector expr max(tmp)]
        } else {
            set min $vec(min)
            set max $vec(max)
        }
        blt::vector destroy tmp zero

        return [list $min $max]
    }

    public method hints {{key ""}} {
        set hints(label) "[$_yvec name] - [$_yvec type]"
        set hints(xlabel) [expr {($xlabel ne "") ? $xlabel : [$_xvec name]}]
        set hints(ylabel) [expr {($ylabel ne "") ? $ylabel : [$_yvec name]}]
        set hints(xscale) $xscale
        set hints(yscale) $yscale
        if {$xlimits ne ""} {
            set hints(xmin) [lindex $xlimits 0]
            set hints(xmax) [lindex $xlimits 1]
        }
        if {$ylimits ne ""} {
            set hints(ymin) [lindex $ylimits 0]
            set hints(ymax) [lindex $ylimits 1]
        }

        if {$key eq ""} {
            return [array get hints]
        } elseif {[info exists hints($key)]} {
            return $hints($key)
        }
        return ""
    }

    public method xErrorValues { cname } { return "" } 
    public method yErrorValues { cname } { return "" } 
    public method xmarkers {} { return "" }
    public method ymarkers {} { return "" }

    private variable _xvec ""
    private variable _yvec ""
    private variable _storage ""  ;# space for the x-vector, if we allocate it
}

# ----------------------------------------------------------------------
# OBJECT: OctaveSession
# Represents a live session that talks to Octave for execution.
# ----------------------------------------------------------------------
itcl::class OctaveSession {
    public variable outputcommand ""
    public variable errorcommand ""
    public variable debug 0

    constructor {args} {
        eval configure $args
    }
    destructor {
        clear
    }

    # discards all variables and restarts Octave
    method clear {} {
        set jobs($this) "kill"
        clearplots
        clearvars
        clearoutput
    }

    # enters a script of commands and waits for results
    method script {script} {
        global InstallDir

        # convert any plot() command to plotoctaview()
        regsub -all {plot *\(} $script "plotoctaview('linlin'," script
        regsub -all {semilogx *\(} $script "plotoctaview('loglin'," script
        regsub -all {semilogy *\(} $script "plotoctaview('linlog'," script
        regsub -all {loglog *\(} $script "plotoctaview('loglog'," script

        # convert any "xlabel"/"ylabel" commands to x/ylabeloctaview()
        regsub -all {xlabel *\(} $script "xlabeloctaview(" script
        regsub -all {ylabel *\(} $script "ylabeloctaview(" script
        regsub -all {title *\(} $script "titleoctaview(" script
        regsub -all {axis *\(} $script "axisoctaview(" script

        # convert any "hold" command to holdoctaview()
        regsub -all {hold +([^ \(;\n]+) *([;\n])} $script "holdoctaview('\\1')\\2" script
        regsub -all {hold *([;\n])} $script "holdoctaview()\\1" script

        # disallow naughty commands
        regsub -all {exec *\(} $script "notallowed(" script
        regsub -all {system *\(} $script "notallowed(" script
        regsub -all {unix *\(} $script "notallowed(" script
        regsub -all {dos *\(} $script "notallowed(" script
        regsub -all {perl *\(} $script "notallowed(" script
        regsub -all {popen *\(} $script "notallowed(" script
        regsub -all {popen2 *\(} $script "notallowed(" script
        regsub -all {fork *\(} $script "notallowed(" script
        regsub -all {pipe *\(} $script "notallowed(" script

        set counter 0
        set scriptfile [file join /tmp "octaview[incr counter]"]
        while {[file exists $scriptfile]} {
            set scriptfile [file join /tmp "octaview[incr counter]"]
        }

        set fid [open $scriptfile w]
        puts -nonewline $fid $script
        close $fid

        set counter 0
        set cmdfile [file join /tmp "octaviewc[incr counter]"]
        while {[file exists $cmdfile]} {
            set cmdfile [file join /tmp "octaviewc[incr counter]"]
        }

        set fid [open $cmdfile w]
        puts $fid "more off; figure(\"visible\",\"off\"); sighup_dumps_octave_core(0);"
        puts $fid "atexit(\"reportvars\");"
        puts $fid "source(\"$scriptfile\");"
        puts $fid "quit;"
        close $fid

        # clear out status -- this will have job termination info later
        set _running 1
        set ::OctaveSession::jobs($this) ""

        catch {blt::bgexec ::OctaveSession::jobs($this) \
            -keepnewline yes \
            -onoutput [itcl::code $this output stdout] \
            -onerror [itcl::code $this output stderr] \
            $octaveprog --path $InstallDir $cmdfile}

        # clean up the tmp files
        catch {file delete -force $cmdfile $scriptfile}
        set _running 0

        # --------------------------------------------------------------
        # Look for debug information ...
        # --------------------------------------------------------------
        if {[regexp -indices -- {---OCTAViEw-REPORTVARS---\n} $_out match]} {
            foreach {s0 s1} $match break
            set info [string range $_out [expr {$s1+1}] end]
            set _out [string range $_out 0 [expr {$s0-1}]]

            while {1} {
                if {[regexp -indices {variable\(([a-zA-Z0-9_]+),([a-zA-Z0-9_]+)\)>>} $info match vname type]} {
                    foreach {s0 s1} $vname break
                    set vname [string range $info $s0 $s1]
                    foreach {s0 s1} $type break
                    set type [string range $info $s0 $s1]
                    foreach {s0 s1} $match break
                    set info [string range $info [expr {$s1+1}] end]
                    set real ""
                    set imag ""

                    set i [string first "<<" $info]
                    if {$i < 0} break

                    set real [string range $info 0 [expr {$i-1}]]
                    set nextpart [string range $info $i [expr {$i+4}]]
                    if {$nextpart eq "<<i>>"} {
                        set info [string range $info [expr {$i+5}] end]
                        set i [string first "<<" $info]
                        if {$i >= 0} {
                            set imag [string range $info 0 [expr {$i-1}]]
                            set info [string range $info [expr {$i+2}] end]
                        }
                    }

                    switch -glob -- $type {
                        double1x* {
                            set obj [VectorResult ::#auto $vname $real $imag]
                            lappend _vars $obj
                        }
                    }

                    # start fresh with the next var
                    set vname ""
                    set vals ""
                } else {
                    # no more variable definitions -- done
                    break
                }
            }
        }

        # --------------------------------------------------------------
        # Look for embedded plots ...
        # --------------------------------------------------------------
        set plothold 0
        while {[regexp -indices -- {--!!plot<<([^|]*)\|([^|]*)\|([^|]*)>>} $_out match yaxis xaxis popts]} {
            foreach {s0 s1} $yaxis break
            set yaxis [string range $_out $s0 $s1]
            foreach {s0 s1} $xaxis break
            set xaxis [string range $_out $s0 $s1]
            foreach {s0 s1} $popts break
            set popts [string range $_out $s0 $s1]

            foreach {s0 s1} $match break
            set info [string range $_out [expr {$s1+1}] end]
            set top [string range $_out 0 [expr {$s0-1}]]
            set btm ""

            # each plot starts fresh unless there's a hold on
            if {!$plothold} {
                set axes ""
            }

            # look for the marker at the end of plotting info
            if {[regexp -indices -- {!!--\n} $info match]} {
                foreach {s0 s1} $match break
                set btm [string range $info [expr {$s1+1}] end]
                set info [string range $info 0 [expr {$s0-1}]]
            }

            # keep all output above/below the --!!plot<<>> stuff
            set _out $top; append _out $btm

            # this may be a "hold" statement
            if {[regexp {^hold=([01])} $popts match state]} {
                set plothold $state
                continue
            }

            # this may be "xlabel" or "ylabel" statements
            if {[regexp {^xlabel=(.+)} $popts match xlabel]} {
                regsub -all {\\} $xlabel "" xlabel
                set last [lindex $_plots end]
                if {$last ne ""} {
                    # update the most recent plot with this label
                    foreach curve [lindex $last 2] {
                        $curve configure -xlabel $xlabel
                    }
                }
                continue
            }
            if {[regexp {^ylabel=(.+)} $popts match ylabel]} {
                regsub -all {\\} $ylabel "" ylabel
                set last [lindex $_plots end]
                if {$last ne ""} {
                    # update the most recent plot with this label
                    foreach curve [lindex $last 2] {
                        $curve configure -ylabel $ylabel
                    }
                }
                continue
            }

            # this may be a "title" statement
            if {[regexp {^title=(.+)} $popts match title]} {
                set last [lindex $_plots end]
                if {$last ne ""} {
                    # update the most recent plot with this title
                    set last [lreplace $last 0 0 $title]
                    set _plots [lreplace $_plots end end $last]
                }
                continue
            }

            # this may be axis limits from an "axis" statement
            if {[regexp {^limits=(.+)} $popts match axlims]} {
                foreach {xmin xmax ymin ymax} [split $axlims ,] break
                set last [lindex $_plots end]
                if {$last ne ""} {
                    # update the most recent plot with this label
                    foreach curve [lindex $last 2] {
                        $curve configure \
                            -xlimits [list $xmin $xmax] \
                            -ylimits [list $ymin $ymax]
                    }
                }
            }

            # peel out axes from front of options
            if {[regexp {^axes=([a-z]+) *(.+)} $popts match axstr rest]} {
                if {$axes eq ""} {
                    set axes $axstr
                }
                set popts $rest
            }

            set vals ""
            set xvals ""
            set yvals ""
            set plabel "plot("
            if {$xaxis ne ""} { append plabel "$xaxis," }
            append plabel "$yaxis)"

            set i [string first "!!x!!" $info]
            if {$i >= 0} {
                set yvals [string range $info 0 [expr {$i-1}]]
                set xvals [string range $info [expr {$i+5}] end]
            } else {
                set yvals $info
                set xvals ""
            }

            if {[llength $yvals] > 0} {
                set yobj [VectorResult ::#auto $yaxis $yvals]
                if {[llength $xvals] > 0} {
                    set xobj [VectorResult ::#auto $xaxis $xvals]
                } else {
                    set xobj ""
                }
                set curve [Curve ::#auto $yobj $xobj]

                switch -- $axes {
                    linlin {
                        # this is the default for curves
                    }
                    linlog {
                        $curve configure -yscale "log"
                    }
                    loglin {
                        $curve configure -xscale "log"
                    }
                    loglog {
                        $curve configure -xscale "log"
                        $curve configure -yscale "log"
                    }
                    default {
                        error "don't understand plot axes \"$axes\""
                    }
                }

                if {$plothold && [llength $_plots] > 0} {
                    set last [lindex $_plots end]

                    # make the label match the first curve, so same axes
                    set curvelist [lindex $last 2]
                    set first [lindex $curvelist 0]
                    $curve configure -xlabel [$first hints xlabel]
                    $curve configure -ylabel [$first hints ylabel]

                    # append this curve to the "curvelist" arg
                    lappend curvelist $curve
                    set last [lreplace $last 2 2 $curvelist]

                    # append the xobj/yobj to the end for cleanup
                    lappend last $yobj $xobj

                    # replace the last plot with this one
                    set _plots [lreplace $_plots end end $last]
                } else {
                    # add a whole new plot to the list
                    lappend _plots [list $plabel $popts $curve $yobj $xobj]
                }
            }
        }

        # --------------------------------------------------------------
        # Look for any errors ...
        # --------------------------------------------------------------
        if {[regexp {parse error (?:near|at) line ([0-9]+) [^\n]+\n([^>]+)>>> [^\n]*\n( +)\^} $_out match line mesg indent]} {
            set mesg [string trim $mesg]
            set col [expr {[string length $indent]-3}]
            _handleError $mesg $line $col
            return 1
        }

        set re [format {error: +(security exception[^\n]+).+ %s at line ([0-9]+),? column ([0-9]+)} $scriptfile]
        if {[regexp $re $_out match mesg line col]} {
            _handleError $mesg $line $col
            return 1
        }
        if {[regexp {error: +([^\n]+) (?:near|at) line ([0-9]+),? +column ([0-9]+)} $_out match mesg line col]} {
            _handleError $mesg $line $col
            return 1
        }

        # no errors -- looks good
        return 0
    }

    method isrunning {} {
        return $_running
    }

    method abort {} {
        set ::OctaveSession::jobs($this) "abort"
    }

    # return the list of plotted results from the Octave session
    method getplots {} {
        return $_plots
    }

    # clear out any plot vectors from a previous run
    method clearplots {} {
        foreach rec $_plots {
            # delete the curves
            foreach obj [lindex $rec 2] {
                itcl::delete object $obj
            }

            # delete all vectors stored within the curves
            foreach vecobj [lrange $rec 3 end] {
                if {$vecobj ne ""} { itcl::delete object $vecobj }
            }
        }
        set _plots ""
    }

    # return the list of current results from the Octave session
    method getvars {} {
        return $_vars
    }

    # clear out any variables from a previous run
    method clearvars {} {
        foreach obj $_vars {
            itcl::delete object $obj
        }
        set _vars ""
    }

    # enters a script of commands and waits for results
    method output {channel str} {
        # save all output for later
        append _out $str

        if {[string length $outputcommand] > 0} {
            # Scan through output and remove plotting info that users
            # aren't expecting
            set clean ""
            set rest $str
            while {1} {
                set found ""
                foreach token {--!!plot<< !!-- ---OCTAViEw-REPORTVARS---} {
                    set i [string first $token $rest]
                    if {$i >= 0} {
                        lappend found [list $i [expr {$i+[string length $token]}] $token]
                    }
                }
                if {$found eq ""} {
                    # no funny tokens left -- keep output unless we're
                    # in the middle of a plotting section
                    if {!$_plotout} {
                        append clean $rest
                    }
                    break
                }

                set first [lindex [lsort -index 0 -integer $found] 0]
                foreach {s0 s1 token} $first break

                switch -- $token {
                    --!!plot<< - ---OCTAViEw-REPORTVARS--- {
                        # starting a plotting section
                        if {!$_plotout} {
                            append clean [string range $rest 0 [expr {$s0-1}]]
                        }
                        set _plotout 1
                        set rest [string range $rest [expr {$s1+1}] end]
                    }
                    !!-- {
                        # ending a plotting section
                        set _plotout 0
                        incr s1  ;# next char after token
                        if {[string index $rest $s1] eq "\n"} {
                            incr s1  ;# skip over trailing newline
                        }
                        set rest [string range $rest $s1 end]
                    }
                }
            }

            # execute the callback but show only clean output
            if {[string length $clean] > 0} {
                uplevel #0 $outputcommand [list $channel $clean]
            }
        }

        if {$debug} {
            puts -nonewline $str
        }
    }

    method getoutput {} {
        return $_out
    }

    method clearoutput {} {
        set _out ""
        set _plotout 0
    }

    # called to invoke the -errorcommand string to report errors in the script
    private method _handleError {err line col} {
        if {$errorcommand ne ""} {
            uplevel #0 $errorcommand [list $err $line $col]
        }
    }

    private variable _running 0 ;# non-zero => script is running
    private variable _vars ""   ;# list of result objects from Octave program
    private variable _plots ""  ;# list of result objects from Octave program
    private variable _out ""    ;# stdout from Octave console
    private variable _plotout 0 ;# non-zero => in the midst of plotting output

    # store the path to the Octave program for later
    common octaveprog [auto_execok octave]
    if {$octaveprog eq ""} {
        tk_messageBox -icon error -message "Can't find the Octave program.  Try \"use octave-3.2.4\" or just \"use\" to see a list of available packages."
        exit
    }
}

# ----------------------------------------------------------------------
# WIDGET: XyResultSet
#
# Bunch of plots and an easy way to switch back and forth between
# them--like the pages of a notebook.  We need this because the
# XyResult widget in Rappture can't really be cleared, but can only
# add more data and hide existing (old) data.  So the only way to
# get it right is to create a separate widget for each plot and clear
# the plot by destroying the widget.
# ----------------------------------------------------------------------
itcl::class XyResultSet {
    inherit itk::Widget

    constructor {args} {
        eval itk_initialize $args
    }
    destructor {
        clear
    }

    public method show {objlist} {
        if {![info exists _curve2win($objlist)]} {
            set n [incr _counter]
            itk_component add plot$n {
                Rappture::XyResult $itk_interior.xy$n
            }
            set _curve2win($objlist) $itk_component(plot$n)

            set c "autoreset"
            foreach obj $objlist {
                $itk_component(plot$n) add $obj [list -color $c]
                set c "auto"
            }
            eval $itk_component(plot$n) scale $objlist
        }

        # clear any other widget and show the current one
        foreach win [pack slaves $itk_interior] {
            pack forget $win
        }
        pack $_curve2win($objlist) -expand yes -fill both
    }

    public method clear {} {
        foreach objlist [array names _curve2win] {
            destroy $_curve2win($objlist)
            unset _curve2win($objlist)
        }
        set _counter 0
    }

    private variable _curve2win   ;# maps curve object => XyResult widget
    private common _counter 0     ;# unique counter for widgets
}

# ----------------------------------------------------------------------
# MAIN WINDOW
# ----------------------------------------------------------------------
frame .in
pack .in -side left -expand yes -fill both -padx 10 -pady 10

frame .in.ctrl
pack .in.ctrl -side top -fill x

if {[Rappture::filexfer::init]} {
    button .in.ctrl.upload -text "Upload..." -command main_file_upload
    Rappture::Tooltip::for .in.ctrl.upload "Upload a file or copy/paste some text from your desktop.\n\nNOTE: You must have pop-ups enabled for this site."
    pack .in.ctrl.upload -side left
    button .in.ctrl.dnload -text "Download..." -command main_file_download
    pack .in.ctrl.dnload -side left -padx {0 20}
    Rappture::Tooltip::for .in.ctrl.dnload "Downlaod the contents of this buffer to your desktop in a pop-up window.\n\nNOTE: You must have pop-ups enabled for this site."
}
button .in.ctrl.open -text "Open..." -command main_file_open
pack .in.ctrl.open -side left
Rappture::Tooltip::for .in.ctrl.open "Load a file from your hub storage area."

button .in.ctrl.save -text "Save As..." -command main_file_saveas
pack .in.ctrl.save -side left
Rappture::Tooltip::for .in.ctrl.save "Save a file to your hub storage area."

Rappture::Scroller .in.scrl -width 2i -height 2i -xscrollmode auto -yscrollmode auto
pack .in.scrl -expand yes -fill both
text .in.scrl.text -wrap none
.in.scrl contents .in.scrl.text

.in.scrl.text tag configure err -background red -foreground white

frame .run
pack .run -side top -fill x -padx 10 -pady {10 0}
frame .run.cntl
pack .run.cntl -fill x
button .run.cntl.go -anchor w -compound left -text "Run Script" -image [Rappture::icon octave-logo-sm] -command main_run
pack .run.cntl.go -side left -expand yes -fill both

label .run.cntl.info -image [Rappture::icon helpinfo] -borderwidth 1 -relief sunken -background gray
Rappture::Tooltip::for .run.cntl.info "Show help information that appears on start-up, including links to tutorials."
pack .run.cntl.info -side right -ipadx 2 -ipady 2

bind .run.cntl.info <ButtonRelease> {main_info -toggle}

frame .right
pack .right -side right -expand yes -fill both -padx 10 -pady 10

# ----------------------------------------------------------------------
# INTRODUCTORY INFO
# ----------------------------------------------------------------------
frame .right.intro
pack .right.intro -expand yes -fill both
Rappture::HTMLviewer .right.intro.html -width 2i
pack .right.intro.html -expand yes -fill both

set fid [open [file join $InstallDir help intro.html] r]
set info [read $fid]
close $fid
.right.intro.html load $info -in [file join $InstallDir help]

# ----------------------------------------------------------------------
# PROGRESS INFO DURING RUN
# ----------------------------------------------------------------------
frame .right.progress
set bg [.right.progress cget -background]
Rappture::Scroller .right.progress.scrl -width 2i -height 2i \
    -xscrollmode auto -yscrollmode auto
pack .right.progress.scrl -expand yes -fill both
text .right.progress.scrl.text -wrap none -state disabled \
    -background $bg -borderwidth 0
.right.progress.scrl contents .right.progress.scrl.text

.right.progress.scrl.text tag configure stderr -foreground red

# ----------------------------------------------------------------------
# PLOTS/RESULTS
# ----------------------------------------------------------------------
frame .right.results
blt::tabset .right.results.tabs -side bottom -tearoff 0
pack .right.results.tabs -expand yes -fill both

set win [frame .right.results.tabs.plots]
.right.results.tabs insert end plots -text "Plots" \
    -window .right.results.tabs.plots -fill both

frame $win.which
pack $win.which -side top -fill x -pady {0 10}
label $win.which.plotl -text "Plot:"
pack $win.which.plotl -side left
Rappture::Combobox $win.which.plot -editable no -width 5
pack $win.which.plot -side left -expand yes -fill x
bind $win.which.plot <<Value>> main_xyplot_update

XyResultSet $win.xy
pack $win.xy -expand yes -fill both


set win [frame .right.results.tabs.debug]
.right.results.tabs insert end debug -text "Debug" \
    -window .right.results.tabs.debug -fill both

frame $win.which
pack $win.which -side top -fill x -pady {0 10}
label $win.which.plotl -text "Plot:"
pack $win.which.plotl -side left
Rappture::Combobox $win.which.plot -editable no -width 5
pack $win.which.plot -side left -expand yes -fill x
frame $win.which.versus
label $win.which.versus.l -text "Versus:"
pack $win.which.versus.l -side left
Rappture::Combobox $win.which.versus.var -editable no -width 5
pack $win.which.versus.var -side left -expand yes -fill x

bind $win.which.plot <<Value>> {main_debug_update yaxis}
bind $win.which.versus.var <<Value>> {main_debug_update xaxis}

XyResultSet $win.xy
pack $win.xy -expand yes -fill both

label $win.other -text "Other Variables:"
pack $win.other -side top -anchor w -pady {12 0}
Rappture::Scroller $win.scrl -width 1i -height 1i -xscrollmode auto -yscrollmode auto
pack $win.scrl -expand yes -fill both
text $win.scrl.info -wrap none
$win.scrl contents $win.scrl.info

# console for Octave output
frame .right.results.tabs.console
.right.results.tabs insert end console -text "Console Output" \
    -window .right.results.tabs.console -fill both
Rappture::Scroller .right.results.tabs.console.scrl -width 2i -height 2i \
    -xscrollmode auto -yscrollmode auto
pack .right.results.tabs.console.scrl -expand yes -fill both
text .right.results.tabs.console.scrl.text -wrap none -state disabled
.right.results.tabs.console.scrl contents .right.results.tabs.console.scrl.text

# use this object to interact with Octave
set Run [OctaveSession ::#auto -debug 0 -errorcommand main_err -outputcommand main_progress]

wm deiconify .
focus .in.scrl.text

if { [info exists env(TOOL_PARAMETERS)] } {
    main_seek_input_from_parameters
} else {
    main_seek_input
}
